//Copyright (c) 2008-2016 Emil Dotchevski and Reverge Studios, Inc.
//Copyright (c) 2018 agate-pris

//Distributed under the Boost Software License, Version 1.0. (See accompanying
//file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_QVM_VEC_REGISTER_HPP
#define BOOST_QVM_VEC_REGISTER_HPP

#include <boost/qvm/vec_register2.hpp>
#include <boost/qvm/vec_register3.hpp>
#include <boost/qvm/vec_register4.hpp>

#endif
