/*//////////////////////////////////////////////////////////////////////////////
    Copyright (c) 2014 Jamboree

    Distributed under the Boost Software License, Version 1.0. (See accompanying
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//////////////////////////////////////////////////////////////////////////////*/
#ifndef FUSION_INCLUDE_FLATTEN
#define FUSION_INCLUDE_FLATTEN


#include <boost/fusion/algorithm/transformation/flatten.hpp>


#endif
