// Copyright 2018 Hans Dembinski
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt
// or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_HISTOGRAM_DETAIL_ATTRIBUTE_HPP
#define BOOST_HISTOGRAM_DETAIL_ATTRIBUTE_HPP

#if __cplusplus >= 201703L
#define BOOST_HISTOGRAM_NODISCARD [[nodiscard]]
#else
#define BOOST_HISTOGRAM_NODISCARD
#endif

#endif
