//  (C) Copyright Gennadiy Rozental 2001.
//  Distributed under the Boost Software License, Version 1.0.
//  (See accompanying file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt)

//  See http://www.boost.org/libs/test for the library home page.
//
/// @file
/// @deprecated
// ***************************************************************************

#include <boost/config/header_deprecated.hpp>
BOOST_HEADER_DEPRECATED( "This header is deprecated. Please use <boost/test/included/unit_test.hpp> instead." )


#include <boost/test/included/unit_test.hpp>
